/**
 * HM acquire - software for reading out HAMEG digital storage
 * oscilloscopes.
 *         
 * Copyright (C) 2007 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * LSto@gmx.de
 * http://formica.nusseis.de/HMacquire/
 */


/*
 * These routines work on pure channel data (double *)
 */

#include "data_api.h"

/**
 * print xy-data to file
 * output can be read by ch_read_xy().
 *
 * f	outpuf file
 * xb	x data to print
 * yb	y data to print
 * xdiv voltage per devision
 * ydiv voltage per devision
 */
void ch_print_xy( FILE * f, const double * xb, const double * yb, 
			 double xdiv, double ydiv )
{
    int i;
    
    fprintf( f, "# %s\n", DATA_MARK );
    fprintf( f, "# X %6.3f V/DIV\n", xdiv );
    fprintf( f, "# Y %6.3f V/DIV\n", ydiv );
    fprintf( f, "# x (V)\ty (V)\n" );
    for ( i = 0; i < 2048; i++ )
	fprintf( f, "%e\t%e\n", xb[i], yb[i] );
}

/**
 * read xy data from file f
 *
 * seeks for DATA_MARK and then reads output from ch_print_xy 
 * data      is stored to xb, yb
 * voltscale is stored to xdiv, ydiv
 * returns 1 on success, 0 on error
 */
int ch_read_xy( FILE * f, double * xb, double * yb, double * xdiv, double * ydiv )
{
  int i;
  char s[1024];
  
  /* seek for DATA_MARK */
  do
    if ( ! fgets( s, 1024, f ) )
      return 0;
  while ( ! strstr( s, DATA_MARK ) );
  /* read header */
  if ( 1 != fscanf( f, "# X %lf V/DIV\n", xdiv ) )
    return 0;
  if ( 1 != fscanf( f, "# Y %lf V/DIV\n", ydiv ) )
    return 0;
  fgets( s, 1024, f ); /* table head with units */
  /* read data */
  for ( i = 0; i < 2048; i++ )
    if ( 1 != fscanf( f, "%le\t%le\n", &xb[i], &yb[i] ) )
      return 0;
  return 1;
}

/**
 * print data for cahannel to file f
 * output can be read by ch_read().
 *
 * f	outpuf file
 * b	data to print
 * tdiv time per division, i.e. per 200 data points
 * ydiv voltage per devision
 */
void ch_print( FILE * f, const double * b, double tdiv, double ydiv )
{
  double time;
  int i;

  fprintf( f, "# %s\n", DATA_MARK );
  fprintf( f, "# Y %6.3f V/DIV\n", ydiv );
  fprintf( f, "# t %6.0e s/DIV\n", tdiv );
  fprintf( f, "# time (s)\tvoltage (V)\n" );
  for ( i = 0; i < 2048; i++ )
  {
    time = ( i - 1023 ) / 200.0 * tdiv;
    fprintf( f, "%e\t%e\n", time, b[i] );
  }
}


/**
 * read data from file f
 *
 * seeks for DATA_MARK and then reads output from ch_print 
 * data      is stored to b
 * timscale  is stored to tdiv
 * voltscale is stored to ydiv
 * returns 1 on success, 0 on error
 */
int ch_read( FILE * f, double * b, double * tdiv, double * ydiv )
{
  int i;
  char s[1024];
  
  /* seek for DATA_MARK */
  do
    if ( ! fgets( s, 1024, f ) )
      return 0;
  while ( ! strstr( s, DATA_MARK ) );
  /* read header */
  if ( 1 != fscanf( f, "# Y %lf V/DIV\n", ydiv ) )
    return 0;
  if ( 1 != fscanf( f, "# t %le s/DIV\n", tdiv ) )
    return 0;
  fgets( s, 1024, f ); /* table head with units */
  /* read data */
  for ( i = 0; i < 2048; i++ )
    if ( 1 != fscanf( f, "%*e\t%le\n", &b[i] ) )
      return 0;
  return 1;
}


/**
 * Add two channels.
 *
 * Destination may be one of the sources.
 *
 * d	destination
 * s1	source 1
 * s2	source2
 */
void ch_add( double * d, const double * s1, const double * s2  )
{
  int i;

  for ( i = 0; i < 2048; i++ )
    d[i] = s1[i] + s2[i];
}


/**
 * Calculate spectrum for channel
 * The values of s are giben in dBV
 *
 * c  channel data, array length 2048
 * s  spectrum, array length 1025
 */
void ch_pwr_spec( const double * c, double * s )
{
  int i;
  #ifdef FFTW2
  double ft[2048];
  double win[2048];
  rfftw_plan p;
  #else
  double * ft; 
  double * win;
  fftw_plan p;
  ft = (double*)fftw_malloc(sizeof(double) * 2048);
  win = (double*)fftw_malloc(sizeof(double) * 2048);
  #endif
  
  /* von Hann window (Hanning) */
  for ( i = 0; i < 2048; i++ )
    win[i] = c[i] * 2.0 * ( 0.5 + 0.5 * cos( 2.0 * M_PI / 2048.0 * i ) );
  /* fft */
  #ifdef FFTW2
  p = rfftw_create_plan( 2048, FFTW_REAL_TO_COMPLEX, FFTW_ESTIMATE );
  rfftw_one( p, win, ft );
  rfftw_destroy_plan( p );
  #else
  p = fftw_plan_r2r_1d( 2048, win, ft, FFTW_R2HC, 
                        FFTW_ESTIMATE | FFTW_DESTROY_INPUT );
  fftw_execute( p );
  fftw_destroy_plan( p );
  #endif
  /* power spectrum */
  s[0] = ft[0] * ft[0];
  for ( i = 1; i < 1024; i++ )
    s[i] = ft[i] * ft[i] + ft[2048-i] * ft[2048-i];
  s[1024] = ft[1024] * ft[1024];
  /* change to dBV */
  for ( i = 0; i < 1025; i++ )
    s[i] = 10 * log( s[i] / 2048 / 2048 ) / log( 10 );
 /* clean up */
 #ifndef FFTW2
 fftw_free( ft );
 fftw_free( win );
 #endif
}


/**
 * print spectrum to file f
 *
 * f	outpuf file
 * b	data to print
 * tdiv time per division, i.e. per 200 data points
 * ydiv voltage per devision
 */
void ch_pwr_spec_print( FILE * f, const double * b, double tdiv, double ydiv )
{
  double freq;
  int i;

  fputs( "# spectrum\n", f );
  fprintf( f, "# Y %6.3f V/DIV\n", ydiv );
  fprintf( f, "# t %6.0e s/DIV\n", tdiv );
  fprintf( f, "# frequency (Hz)\tLevel (dBV)\n" );
  for ( i = 0; i < 1025; i++ )
  {
    freq = i * 200.0 / tdiv / 2048.0;
    fprintf( f, "%e\t%e\n", freq, b[i] );
  }
}


/** 
 * Re[ ( a + ib ) ( c - id ) ]
 */
#define Re_mul_c(a, b, c, d) ((a)*(c)+(b)*(d))


/** 
 * Im[ ( a + ib ) ( c - id ) ]
 */
#define Im_mul_c(a, b, c, d) ((b)*(c)-(a)*(d))


/**
 * Calculate normalised correlation function of f and g.
 *
 * d	destination, length = 4096
 * s1	source 1, length = 2048
 * s2	source 2, length = 2048
 */
void ch_corr( double * d, const double * s1, const double * s2  )
{
  int i;
  double re, im, dc;
  double * ft1; 
  double * ft2; 
  fftw_plan pf1;
  fftw_plan pf2;
  fftw_plan pb;
  ft1 = (double*)fftw_malloc(sizeof(double) * 4096);
  ft2 = (double*)fftw_malloc(sizeof(double) * 4096);
  
  /* fftw init */
  pf1 = fftw_plan_r2r_1d( 4096, ft1, ft1, FFTW_R2HC, FFTW_ESTIMATE );
  pf2 = fftw_plan_r2r_1d( 4096, ft2, ft2, FFTW_R2HC, FFTW_ESTIMATE );
  pb = fftw_plan_r2r_1d( 4096, ft1, d, FFTW_HC2R,
                         FFTW_ESTIMATE | FFTW_DESTROY_INPUT );
  /* copy input to fft buffer; 
   * contininous convolution: pad with signal zero (signal DC) */
  dc = 0;
  for ( i = 0; i < 2048; i++ )
    dc += ft1[i] = s1[i];
  for ( i = 2048; i < 4096; i++ )
    ft1[i] = dc / 2048;
  dc = 0;
  for ( i = 0; i < 2048; i++ )
    dc += ft2[i] = s2[i];
  for ( i = 2048; i < 4096; i++ )
    ft2[i] = dc / 2048;
  /* forward transforms */
  fftw_execute( pf1 );
  fftw_execute( pf2 );
  /* F(w) x G*(w) */
  ft1[0] = 0; /* eliminate DC */
  for ( i = 1; i < 2048; i++ )
  {
    re = Re_mul_c( ft1[i], ft1[4096-i], ft2[i], ft2[4096-i] );
    im = Im_mul_c( ft1[i], ft1[4096-i], ft2[i], ft2[4096-i] );
    ft1[i]      = re;
    ft1[4096-i] = im;
  }
  ft1[2048] = ft1[2048] * ft2[2048];
  /* backward transform */
  fftw_execute( pb );
  /* normalise: FFTW does not normalise, i.e. it multiplies by "duration" N.
   * As signal length is only half of FFT length, multiply by 2. */
  for ( i = 0; i < 4096; i++ )
    d[i] = d[i] / 4096 / 4096 * 2;
  /* clean up */
  fftw_destroy_plan( pf1 );
  fftw_destroy_plan( pf2 );
  fftw_destroy_plan( pb );
  fftw_free( ft1 );
  fftw_free( ft2 );
}



/**
 * print correlation to file f
 *
 * f	outpuf file
 * b	data to print
 * tdiv time per division, i.e. per 200 data points
 */
void ch_corr_print( FILE * f, const double * b, double tdiv, double ydiv )
{
  double time;
  int i;

  fprintf( f, "# Y %6.3e V/DIV\n", ydiv );
  fprintf( f, "# t %6.0e s/DIV\n", tdiv );
  fprintf( f, "# time (s)\tcorrelation\n" );
  for ( i = 2048; i < 4096; i++ )
  {
    time = ( i - 4096 ) / 200.0 * tdiv;
    fprintf( f, "%e\t%e\n", time, b[i] );
  }
  for ( i = 0; i < 2048; i++ )
  {
    time = i / 200.0 * tdiv;
    fprintf( f, "%e\t%e\n", time, b[i] );
  }
}

