/**
 * HMLCR - software to measure inductance and capacitance with a
 * Hameg digital storage oscilloscope
 *         
 * Copyright (C) 2015 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * formica@opppf.de
 * http://opppf.de/
 */
#define _ISOC99_SOURCE
#include <math.h>
#include <complex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <termios.h>
#include "osc_api.h"
#include "si_prefix.h"

#define VERSION "1.02 2017-08-22"
#define SERIAL_PORT "/dev/ttyS1"
#define ALSA_DEVICE "default"
#define R_DEFAULT 1000.0

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

/* debugging */
#define debugc(x) fprintf( stderr, "%s=%c\n", #x, (x) )
#define debugi(x) fprintf( stderr, "%s=%d   %#06x\n", #x, (x), (x) )
#define debugd(x) fprintf( stderr, "%s=%g\n", #x, (x) )
#define debugz(x) fprintf( stderr, "%s=%g %+g i\n", #x, creal(x), cimag(x) )
#define debugp(x) fprintf( stderr, "%s=%g exp(i %g)\n", #x, cabs(x), carg(x) )

/* ---------------------------------------------------------------------- 
 * typedefs
 * ---------------------------------------------------------------------- */

#if OSC_ERROR_STR_LEN > 256
#define ERRORLEN OSC_ERROR_STR_LEN
#else
#define ERRORLEN 256
#endif

/** command line args */
typedef struct
{
   int help;	    /* flag -h */
   int version;	    /* flag -v */
   int interactive; /* flag -i */
   int autoscale;   /* flag -a */
   char * serial;   /* -s <serial port> */
   double R;	    /* -r current measurement resistor */
   double f;	    /* -f measurement frequency */
   char wiring;	    /* -S Series, -Y sYmmetric */
   char printing;   /* -H human, -X exponential */
   complex _r_p;  /* 1/R_p, paralle impedance of resistor R */
   complex _z_p;  /* 1/Z_p, paralle impedance of dut Z */
} t_args;

/** scope data */
typedef struct
{
    int Yi1, Yi2, Ti;	    /* voltage and time scale as index */
    int Yvar1, Yvar2, Tvar; /* voltage and time variable scale setting */
    int state1, state2;	    /* channel settings (osc_get_ch_state) */
    double ch1[2048];	    /* data channel 1 */ 
    double ch2[2048];	    /* data channel 2 */ 
    char error[ERRORLEN];   /* 0-terminated, length 0 means no error */
} t_scope_data;

/** signal data */
typedef struct
{
    double f;		    /* measurement frequency */
    complex Ur_U;	    /* voltage ratio U_r/U */
    char error[ERRORLEN];   /* 0-terminated, length 0 means no error */
} t_signal;

/** parameters for Device Unter Test */
typedef struct
{
    complex Z;    /* Impedance */
    double f;	    /* test frequency */
    double val;	    /* imaginary part as inductance */
    int is_L;	    /* flag: inductive (1), capacitive (0) */
} t_DUT;

/** calibration measurement */
typedef struct
{
    complex Ur_U_open; /* voltage ratio Ur/U for Z = open */
    complex Ur_U_ref;  /* voltage ratio Ur/U for Z = R */
    int open;		 /* flag: Ur_U_open is set */
    int ref;		 /* flag: Ur_U_ref is set */
} t_calibration;

/* ---------------------------------------------------------------------- 
 * little helpers
 * ---------------------------------------------------------------------- */

/** Print error message and exit with code 1 */
void error( const char * e )
{
  fputs( e, stderr );
  fputs( "\nHelp with -h.\n", stderr );
  exit( 1 );
}

/** Print information on current line */
void inform( const char * e )
{
     fprintf( stderr, "%s\n", e );
}

/** maximum(x,y) */
#define minimum(x,y) (((x)<(y))?(x):(y))

/** minimum(x,y) */
#define maximum(x,y) (((x)>(y))?(x):(y))

/* ---------------------------------------------------------------------- 
 * channel math
 * ---------------------------------------------------------------------- */

/**
 * debug: print buffer data
 *
 * f		filehandle to print to
 * b		buffer
 * N		number of samples
 */
void ch_print( FILE * f, const double * b, int N )
{
    int i;

    fputs( "\n\n", f );
    for ( i = 0; i < N; i++ )
	fprintf( f, "%d %g\n", i, b[i] );
}

/**
 * invert all values in buffer: *(-1)
 *
 * b		buffer
 * N		number of values
 */
void ch_invert( double * b, int N )
{
    int i;
    for ( i = 0; i < N; i++ )
	b[i] *= -1;
}

/**
 * determine complex amplitude of given frequency
 *
 * Definition of amplitude Û and phase φ:
 *   U = Û sin( 2πft + φ )
 *
 * b		buffer
 * Ns		number of samples per buffer
 * Nw		number of whole waves in each buffers
 * return: complex amplitude
 */
complex ch_c_amp( const double * b, int Ns, int Nw )
{
    int i;
    complex A;

    A = 0;
    for ( i = 0; i < Ns; i++ )
	A += b[i] * cexp( -I * 2 * M_PI * Nw / Ns * i );
    /* complex amplitudes, engineer's definition: 
     *	    sin(2πft+φ) = sin(ωt+φ)
     *
     * DFT-coefficient:
     *      Σ sin(ωt+φ) * exp(-iωt)
     *    = Σ -i/2 [ exp(iωt+iφ) - exp(-iωt-iφ) ] exp(-iωt)   with -i=exp(-iπ/2)
     *    = ½ Σ exp(iφ-iπ/2) - exp(-i2ωt-iφ-iπ/2)
     *        '------------'   '----------------'
     *           n times       sum over whole periode -> 0
     *    = N/2 exp(iφ-iπ/2)
     *
     * => amplitude normalization: N/2
     * => argument normalization:  argDFT = φ-π/2   =>   φ = argDFT + π/2
     */
    A /= (double)Ns/2.0;
    A *= cexp( I*M_PI/2 );
    return A;
}

#define REL_HYST 0.1
/**
 * determine periode length in samples
 * 
 * b		buffer 
 * N		number of samples in buffer
 * return: periode length in samples
 *	   -1 if frequency is not determinable
 *  TODO: besser bei Rauschen
 */
double ch_periode( const double * b, int N )
{
    int i;
    double min, max;
    double hysteresis, th, th_up, th_down;
    int schmitt, first_schmitt; /* 0 undef, 1 above, -1 below */
    int Nw, first, last;
    double T;

    /* determine min and max */
    min = max = b[0];
    for ( i = 0; i < N; i++ )
    {
	max = maximum( b[i], max );
	min = minimum( b[i], min );
    }
    /* determine thesholds, init schmitt-trigger */
    hysteresis = REL_HYST * (max - min);
    if ( b[0] > max - hysteresis )
    {
	th = max - hysteresis;
	schmitt = 1;
    }
    else if ( b[0] < min + hysteresis )
    {
	th = min + hysteresis;
	schmitt = -1;
    }
    else
    {
	th = b[0];
	schmitt = 0;
    }
    th_up = th + 0.5*hysteresis;
    th_down = th - 0.5*hysteresis;
    /* count waves */
    Nw = 0;
    first_schmitt = 0;
    first = -1; /* for -Wall */
    for ( i = 0; i < N; i++ )
    {
	if ( ( b[i] > th_up && schmitt != 1 )
	   ||( b[i] < th_down && schmitt != -1 ) )
	{
	    schmitt = (b[i]>th_up) ? 1 : -1;
	    if ( first_schmitt == 0 )
	    {
		first_schmitt = schmitt;
		first = i;
	    }
	    else if ( first_schmitt == schmitt )
	    {
		last = i;
		Nw++;
	    }
	}
    }
    /* calculate periode length */
    if ( Nw > 0 )
	T = ( (double)last - (double)first ) / (double)Nw;
    else 
	T = -1;
    return T;
}

/* ---------------------------------------------------------------------- 
 * measurement and evaluation
 * ---------------------------------------------------------------------- */

/* Model of impedance measurement
{
 *
 *	    power supply transformer
 *		     UUUUUUU 
 * -U_R, Ch2         =======            U, Ch1
 * o-----+-----+-----UUUUUUU-----+-----+-----o
 *       |     |                 |     |
 *       |     |                 |     |
 *     .-+-. .-+-.             .-+-. .-+-.
 *     |   | |   |             |   | |   |
 *     |R_p| | R |             | Z | |Z_p|
 *     |   | |   |             |   | |   |
 *     '-+-' '-+-'             '-+-' '-+-'
 *       |     |                 |     |
 *       |     |                 |     |
 *       '-----+--------+--------+-----'
 *                      |
 *                    =====
 *
 * Z		impedance to measure
 * R		current measurement resistor
 * Z_p, R_p	parasitic parallel impedances
 *
 * Determine impedance Z: 
 * signal_to_dut()
 *
 *     Z||Z_p = U / U_r ( R || R_p )
 *
 *         [ U_R /  1     1  \    1  ] -1
 * =>  Z = [ --- | --- + --- | - --- ]
 *         [  U  \  R    R_p /   Z_p ]
 *
 * Determine parasitic parallel impedances R_p and Z_p:
 * calibrate()
 *
 *     open: measurement with Z open; 1/Z = 0
 *
 *        1    U_R /  1     1  \                     U_R
 *       --- = --- | --- + --- |            with a = ---
 *       Z_p    U  \  R    R_p /                      U
 *
 *     reference: measurement with Z = R:
 *
 *        1     1    U_R /  1     1  \               U_R
 *       --- + --- = --- | --- + --- |	    with b = ---
 *        R    Z_p    U  \  R    R_p /                U
}
 */

/**
 * Read data from scope
 *
 * serial	device name of serial port
 * autoscale	boolean: perform autoscale before measurement
 * return: scope data
 */
t_scope_data measure( const char * serial, int autoscale )
{
    OSC * o;
    t_scope_data sd;
    int ret;

    /* init */
    o=osc_open( serial );
    if ( o->error[0] ) 
    {
	strcpy( sd.error, "Can not open serial port for scope." );
	goto clean_exit;
    }
    if ( osc_is_ct(o) ) 
    {
	strcpy( sd.error, "Scope is in component test mode (CT)." );
	goto clean_exit;
    }
    if ( osc_is_xy(o) ) 
    {
	strcpy( sd.error, "Scope is in xy-mode." );
	goto clean_exit;
    }
    if ( !osc_is_store(o) ) 
    {
	strcpy( sd.error, "Scope must be in store mode." );
	goto clean_exit;
    }
    /* get scope settings */
    inform( ".. Getting scope settings" );
    if ( autoscale ) 
    {
	osc_set_auto( o );
	sleep( 1 );
    }
    osc_set_hold( o, 1 );
    sd.state1 = osc_get_ch_state( o, 1 );
    sd.state2 = osc_get_ch_state( o, 2 );
    sd.Yi1 = osc_get_ch_scale( o, 1, &sd.Yvar1 );
    sd.Yi2 = osc_get_ch_scale( o, 2, &sd.Yvar2 );
    sd.Ti = osc_get_tb_scale( o, 'A', &sd.Tvar );
    if ( ! osc_ch_on(o,1) ) 
    {
	strcpy( sd.error, "Channel 1 is off." );
	goto clean_exit;
    }
    if ( ! osc_ch_on(o,2) ) 
    {
	strcpy( sd.error, "Channel 2 is off." );
	goto clean_exit;
    }
    if ( osc_ch_gnd(o,1) ) inform( "Channel 1 is grounded (GND)." );
    if ( osc_ch_gnd(o,2) ) inform( "Channel 2 is grounded (GND)." );
    if ( sd.Yvar1 != 0xff )
	inform( "Voltage scale of channel 1 is not calibrated." );
    if ( sd.Yvar2 != 0xff )
	inform( "Voltage scale of channel 2 is not calibrated." );
    if ( sd.Tvar != 0 )
	inform( "Timebase A is not calibrated." );
    /* read data from scope */
    inform( ".. Reading channel 1" );
    ret = osc_read_chan( o, 1, 0, 2048, sd.ch1 );
    switch ( ret )
    {
	case -1:
	    strcpy( sd.error, o->error );
	    goto clean_exit;
	case -2:
	    strcpy( sd.error, "Channel 1 clips." );
	    goto clean_exit;
    }
    if ( osc_ch_inv(o,1) )
	ch_invert( sd.ch1, 2048 );
    inform( ".. Reading channel 2" );
    ret = osc_read_chan( o, 2, 0, 2048, sd.ch2 );
    switch ( ret )
    {
	case -1:
	    strcpy( sd.error, o->error );
	    goto clean_exit;
	case -2:
	    strcpy( sd.error, "Channel 2 clips." );
	    goto clean_exit;
    }
    if ( osc_ch_inv(o,2) )
	ch_invert( sd.ch2, 2048 );
    /* return with no error */
    sd.error[0] = 0; 
clean_exit:
    /* clean up */
    osc_set_hold( o, 0 );
    osc_close( o );
    return sd;
}

/**
 * extract signal properties from scope data
 *
 * sd	    data from scope
 * f	    measurement frequency; 0 means determine from data
 * return: signal properties
 */
t_signal extract_signal( t_scope_data sd, double f )
{
    t_signal signal;
    int Nw;	    /* Number of whole waves in buffer */
    int Ns;	    /* Number of samples containing whole number of waves */
    double T;	    /* periode length in samples */
    complex A1, A2;

    /* clean up data */
    ch_invert( sd.ch2, 2048 );
    /* frequency, periode length, number of waves */
    if ( f )
    {
	/* determine periode length from frequency */
	signal.f = f;
	T = 2000.0 / (10.0 * OSC_s_DIV[sd.Ti]) / signal.f;
    }
    else
    {
	/* determine frequency and periode length from scope data */
	T = ch_periode( sd.ch1, 2048 );
	if ( T < 0 )
	{
	    strcpy( signal.error, "Can't determine frequency." );
	    return signal;
	}
	signal.f = 2000.0 / (10.0 * OSC_s_DIV[sd.Ti]) / T;
    }
    Nw = (int)floor( 2048 / T );
    Ns = (int)floor( Nw * T );
    /* determine complex voltage ratio U_r/U */
    A1 = ch_c_amp( sd.ch1, Ns, Nw );
    A2 = ch_c_amp( sd.ch2, Ns, Nw );
    signal.Ur_U = A2 / A1;
    signal.error[0] = 0;
    return signal;
}

/**
 * calc DUT parameters from signal properties
 *
 * s	    signal properties
 * R	    current resistor R
 * a	    voltage ratio Ur/U for Z = open
 * b	    voltage ratio Ur/U for Z = R
 * return: DUT parameters 
 */
t_DUT signal_to_dut( t_signal s, double R, complex _r_p, complex _z_p )
{
    t_DUT dut;

    /* impedance Z
     *
     *     / U_r /  1     1  \    1  \-1   /      /  1         \        \-1
     * Z = | --- | --- + --- | - --- |   = | Ur_U | --- + _r_p | - _z_p |
     *     \  U  \  R    R_p /   Z_p /     \      \  R         /        /
     */
    dut.Z = 1.0 / ( s.Ur_U * ( 1.0/R + _r_p ) - _z_p );
    /* test frequency f and device values */
    dut.f = s.f;
    if ( cimag(dut.Z) > 0 )
    {
	dut.val = cimag(dut.Z) / 2.0 / M_PI / dut.f;
	dut.is_L = 1;
    }
    else
    {
	dut.val = -1.0 / cimag(dut.Z) / 2.0 / M_PI / dut.f;
	dut.is_L = 0;
    }
    return dut;
}

/**
 * calculate parallel impedances from voltage ratios
 *
 * cal	    calibration data (voltage ratios)
 * arg	    arguments, containing invers parallel impedances
 * return: done in arg
 */
void calibrate( t_calibration cal, t_args * arg )
{
    complex a, b;

    a = cal.Ur_U_open;
    b = cal.Ur_U_ref;
    /* parasitic impedance R_p 
     *
     *         1     1   b - a - 1
     * _r_p = --- = --- -----------
     *        R_p    R     a - b
     */
    arg->_r_p = (b - a - 1)/(a - b)/arg->R;
    /* parasitic impedance Z_p 
     *
     *         1      /  1     1  \     /  1         \
     * _z_p = --- = a | --- + --- | = a | --- + _r_p |
     *        Z_p     \  R    R_p /     \  R         /
     */
    arg->_z_p = a * ( 1.0/arg->R + arg->_r_p );
}

/* ---------------------------------------------------------------------- 
 * printing and formatting
 * ---------------------------------------------------------------------- */

/** Print impedance, human readable */
void print_impedance_human( t_DUT dut )
{
    t_si_prefix p1, p2;
    complex Y;

    /*          1         2         3         4
     * 1234567890123456789012345678901234567890
     *         T       T       T       T       T
     * R  = -0.00 MOhm         G = -0.00 MS
     * phi=-180.0 °
     * Q  =1000.0              d =  0.01
     */
    Y = 1.0/dut.Z;
    p1 = si_prefix( dut.f, 3 );
    printf( "test condition: f = %s\n", format_si( dut.f, p1, "Hz" ) );
    p1 = si_prefix( cabs(dut.Z), 3 );
    p2 = si_prefix( cabs(Y), 3 );
    printf( "R  = %s\t\t", format_si( creal(dut.Z), p1, "Ohm" ) );
    printf( "G  = %s\n", format_si( creal(Y), p2, "S" ) );
    printf( "X  = %s\t\t", format_si( cimag(dut.Z), p1, "Ohm" ) );
    printf( "B  = %s\n", format_si( cimag(Y), p2, "S" ) );
    printf( "Z  = %s\t\t", format_si( cabs(dut.Z), p1, "Ohm" ) );
    printf( "Y  = %s\n", format_si( cabs(Y), p2, "S" ) );
    printf( "phi=%6.1f °\n", 180.0 / M_PI * carg(dut.Z) ); 
    printf( "Q  =%6.1f\t\t", fabs(cimag(dut.Z)/creal(dut.Z)) ); 
    printf( "d  =%6.2f\n", fabs(creal(dut.Z)/cimag(dut.Z)) );
    p1 = si_prefix( dut.val, 3 );
    if ( dut.is_L )
	printf( "L  = %s\n", format_si( dut.val, p1, "H" ) );
    else
	printf( "C  = %s\n", format_si( dut.val, p1, "F" ) );
}

/** Print impedance, exponential notation */
void print_impedance_exp( t_DUT dut )
{
    complex Y;

    Y = 1.0/dut.Z;
    printf( "f    [Hz] = %10.4g\n", dut.f );
    printf( "R   [Ohm] = %10.4g\n", creal(dut.Z) );
    printf( "X   [Ohm] = %10.4g\n", cimag(dut.Z) );
    printf( "Z   [Ohm] = %10.4g\n", cabs(dut.Z) );
    printf( "G     [S] = %10.4g\n", creal(Y) );
    printf( "B     [S] = %10.4g\n", cimag(Y) );
    printf( "Y     [S] = %10.4g\n", cabs(Y) );
    printf( "phi   [°] = %10.4g\n", 180.0/M_PI*carg(dut.Z) );
    printf( "Q         = %10.4g\n", fabs(cimag(dut.Z)/creal(dut.Z)) );
    printf( "d         = %10.4g\n", fabs(creal(dut.Z)/cimag(dut.Z)) );
    if ( dut.is_L )
	printf( "L   [H]   = %10.4g\n", dut.val );
    else
	printf( "C   [F]   = %10.4g\n", dut.val );
}

/* ---------------------------------------------------------------------- 
 * interactive mode
 * ---------------------------------------------------------------------- */

/* buffer length for strings used in interactive mode */
#define INTBUFLEN 256

/**
 * reads one char unbuffered from stdin
 *
 * returns -1 on error
 */
char readchar()
{
  char ch;
  int n;
  static struct termios Otty, Ntty;

  fflush( stdout );
  tcgetattr( 0, &Otty );
  Ntty = Otty;
  /* unbuffered settings */
  Ntty.c_iflag = 0;
  Ntty.c_oflag = 0;
  Ntty.c_lflag &= ~ICANON; /* switch off line buffer */
  /*Ntty.c_lflag |= ECHO;  * enable echo */
  Ntty.c_lflag &= ~ECHO;  /* disable echo */
  Ntty.c_cc[VMIN]  = 1; /* minimum chars to wait for */
  Ntty.c_cc[VTIME] = 255; /* minimum wait time (ds). Does value matter? */
  if ( 0 != tcsetattr( 0, TCSAFLUSH, &Ntty) )
    return -1;
  n = read( 0, &ch, 1 ); 
  if ( 0 != tcsetattr( 0, TCSAFLUSH, &Otty) )
    return -1;
  if ( n != 1 )
    return -1;
  return (int)ch;
}

/**
 * read a line from f to buffer b with length l
 * If string in b ends on NL, replace it by \0.
 */
char * readline( FILE * f, char * b, size_t l )
{
  char * r;
  size_t n;

  r = fgets( b, l, f );
  if ( r ) 
  {
    n = strlen( b ) - 1;
    if ( b[n] == '\n') 
      b[n] = '\0';
  }
  return r;
}

/**
 * Ask for string
 *
 * give prompt "q [buf]> "
 * On empty input buffer is unchanged.
 * assumed buffer length is INTBUFLEN.
 *
 * q    question
 * buf  buffer to read to
 */
void ask_str( const char * q, char * buf )
{
  char line[INTBUFLEN] = "";

  fprintf( stdout, "%s [%s]> ", q, buf );
  fflush( stdout );
  readline( stdin, line, INTBUFLEN );
  if ( line[0] )
    strcpy( buf, line );
}

/**
 * print formatted button, with braces if selected
 *
 * s	    string on button
 * a,b	    if equal, print selected button
 * return: pointer to static formatted string
 */
void button( const char * s, int a, int b )
{
    if ( a == b )
	printf( "\e[1m[%s]\e[m", s ); /* ANSI ESC */
    else
	printf( "|%s|", s ); /* ANSI ESC */
}

/**
 * Print promt for interactice mode
 */
void prompt( t_args * arg, const char * m, t_calibration cal, t_DUT dut )
{
    t_si_prefix p;

    fputs( "\e[H\e[J", stdout ); /* clear screen */
    printf( "\e[7m%s\e[m\n", m );
    /* impedance */
    if ( arg->printing == 'H' )
	print_impedance_human( dut );
    else
	print_impedance_exp( dut );
    printf( "\n" );
    /* line 1 */
    printf( "display    : " );
    button( "Human", arg->printing, 'H' );
    printf( "       " );
    button( "eXponential", arg->printing, 'X' );
    printf( "\n" );
    /* line 2 */
    printf( "calibrate  : Open(%c)  rEference(%c) Clear\n",
	     cal.open?'x':' ', cal.ref?'x':' ' );
    /* line 3 */
    printf( "measurement: " );
    button( "Autoscale", arg->autoscale, 1 );
    printf( "\n" );
    /* line 4 */
    p = si_prefix( arg->R, 3 );
    printf( "             R = %s\n", format_si( arg->R, p, "Ohm" ) );
    /* line 5 */
    if ( arg->f )
    {
	p = si_prefix( arg->f, 3 );
	printf( "             F = %s\n", format_si( arg->f, p, "Hz" ) );
    }
    else
    {
	printf( "             F =  auto\n" );
    }
    /* line 6 */
    printf( "Quit                          <-'measure\n" );

    //print the complex parallel conductivity
    //printf( "_r_p = %g + i %g\n", creal(arg->_r_p), cimag(arg->_r_p) );
    //printf( "_z_p = %g + i %g\n", creal(arg->_z_p), cimag(arg->_z_p) );
    //fprintf( "         1          2        3         4\n" );
    //fprintf( "1234567890123456789012345678901234567890\n" );
}

/**
 * interactice mode
 */
void interactive( t_args * arg )
{
    int c;
    char buf[INTBUFLEN];
    char message[INTBUFLEN];
    t_scope_data sd;
    t_signal signal;
    t_calibration cal;
    t_DUT dut;

    message[0] = 0;
    cal.open = cal.ref = 0;
    dut.Z = 0;
    dut.f = dut.val = dut.is_L = 0;
    do
    {
	prompt( arg, message, cal, dut );
	c = readchar();
	switch ( c )
	{
	    case 'a':
		arg->autoscale = 1 - arg->autoscale;
		message[0] = 0;
		break;
	    case 'c':
		arg->_r_p = 0;
		arg->_z_p = 0;
		cal.open = 0;
		cal.ref = 0;
		message[0] = 0;
		break;
	    case 'e':
		sd = measure( arg->serial, arg->autoscale );
		if ( sd.error[0] )
		{
		    strcpy( message, sd.error );
		    break;
		}
		signal = extract_signal( sd, arg->f );
		if ( signal.error[0] )
		{
		    strcpy( message, signal.error );
		    break;
		}
		cal.Ur_U_ref = signal.Ur_U;
		cal.ref = 1;
		if ( cal.open )
		    calibrate( cal, arg );
		message[0] = 0;
		break;
	    case 'f':
		message[0] = 0;
		sprintf( buf, "%7.4g", arg->f );
		ask_str( "\nfrequency f (0=auto):", buf );
		arg->f = scan_si( buf );
		if ( arg->f < 0 )
		{
		    arg->f = 0;
		    strcpy( message, "f must not be negative." );
		}
		break;
	    case 'h':
		arg->printing = 'H';
		message[0] = 0;
		break;
	    case 'o':
		sd = measure( arg->serial, arg->autoscale );
		if ( sd.error[0] )
		{
		    strcpy( message, sd.error );
		    break;
		}
		signal = extract_signal( sd, arg->f );
		if ( signal.error[0] )
		{
		    strcpy( message, signal.error );
		    break;
		}
		cal.Ur_U_open = signal.Ur_U;
		cal.open = 1;
		if ( cal.ref )
		    calibrate( cal, arg );
		message[0] = 0;
		break;
	    case 'r':
		message[0] = 0;
		sprintf( buf, "%7.4g", arg->R );
		ask_str( "\nshunt resistor R:", buf );
		arg->R = scan_si( buf );
		if ( arg->R <= 0 )
		{
		    arg->R = R_DEFAULT;
		    strcpy( message, "R must be positive." );
		}
		break;
	    case 'x':
		arg->printing = 'X';
		message[0] = 0;
		break;
	    case '\r':
	    case '\n':
		sd = measure( arg->serial, arg->autoscale );
		if ( sd.error[0] )
		{
		    strcpy( message, sd.error );
		    break;
		}
		signal = extract_signal( sd, arg->f );
		if ( signal.error[0] )
		{
		    strcpy( message, signal.error );
		    break;
		}
		dut = signal_to_dut( signal, arg->R, arg->_r_p, arg->_z_p );
		message[0] = 0;
		break;
	}
    } while ( c != 'q' );
}

/* ---------------------------------------------------------------------- 
 * main program and helpers
 * ---------------------------------------------------------------------- */

/** Print help to stdout and exit with 0 */
void help()
{
    printf(
        "HMlcr\n"
        "(c) Lars Stollenwerk, 2015\n"
        "\n"
        "hmlcr -h          print this help\n"
        "hmlcr -v          print version and copyright\n"
        "hmlcr -v serial   print scope firmware version\n"
        "hmlcr [options]   measure impedance\n"
        "\n"
        "options:\n"
        "  -s serial         serial port, default '%s'\n"
        "  -i                interactive mode\n"
        "  -r                resistor for current measurement, default %g Ohm\n"
        "  -f                measurement frequency, 0 = auto (default)\n"
        "  -a                make autoscale before measurement\n"
        "  output\n"
        "  -H                print human readable results (default)\n"
        "  -X                print results in exponential notation \n"
        "\n"
        "                        ___                        \n"
        "               .-----.   |   .-----.               \n"
        "      Ch2 o-o--|  R  |---o---|  Z  |--o-o Ch1      \n"
        "            |  '-----'       '-----'  |            \n"
        "            |                         |            \n"
        "            '---------uuuuuuu---------'            \n"
        "                  .---uuuuuuu---.                  \n"
        "                  |             |                  \n"
        "                  '---o U_in o--'                  \n"
        "\n", 
	SERIAL_PORT, R_DEFAULT );
    exit( 0 );
}

/** Print program version and copyright and exit with 0 */
void copyleft()
{
  puts( "HMLCR - measuring inductance and capacitance" );
  puts( "        using a HAMEG oscilloscope" );
  puts( "" );
  printf( "Version %s\n", VERSION );
  puts( "Copyright (C) 2015 Lars Stollenwerk" );
  puts( "" );
  puts( "This program is free software; you can redistribute it and/or" );
  puts( "modify it under the terms of the GNU General Public License as" );
  puts( "published by the Free Software Foundation; either version 2 of" );
  puts( "the License, or (at your option) any later version." );
  puts( "" );
  puts( "This program is distributed in the hope that it will be" );
  puts( "useful, but WITHOUT ANY WARRANTY; without even the implied" );
  puts( "warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR" );
  puts( "PURPOSE. See the GNU General Public License for more" );
  puts( "details." );
  puts( "" );
  puts( "You should have received a copy of the GNU General" );
  puts( "Public License along with this program; if not, write" );
  puts( "to the Free Software Foundation, Inc., 51 Franklin St," );
  puts( "Fifth Floor, Boston, MA 02110, USA" );
  puts( "" );
  puts( "Contact the author via" );
  puts( "formica@opppf.de" );
  puts( "http://opppf.de/" );
  exit( 0 );
}

/** Print scope version to stderr */
void version( const char * port )
{
  OSC * osc;

  osc = osc_open( port );
  if ( osc->error[0] ) 
  {
    fprintf( stderr, "Error: %s\n", osc->error );
    exit( 1 );
  }
  else
    fprintf( stderr, "Oscilloscope Version: %s\n", osc->version );
  osc_close( osc );
  exit( 0 );
}

/** read commandline args */
void read_args( int argc, char ** argv, t_args * arg )
{
    int i;

    /* set defaults */
    arg->help = 0;
    arg->version = 0;
    arg->interactive = 0;
    arg->autoscale = 0;
    arg->wiring = 'Y';
    arg->printing = 'H';
    arg->serial = SERIAL_PORT;
    arg->R = R_DEFAULT;
    arg->f = 0;
    arg->_r_p = 0;
    arg->_z_p = 0;

    /* options, beginning with - */
    i = 0;
    while ( ++i < argc && argv[i][0] == '-' )
    {
	switch ( argv[i][1] )
	{
	    /* flags */
	    case 'h': arg->help = 1; break;
	    case 'i': arg->interactive = 1; break;
	    case 'a': arg->autoscale = 1; break;
	    /* choices */
	    case 'H': arg->printing = 'H'; break;
	    case 'X': arg->printing = 'X'; break;
	    /* parameters */
	    case 'f': 
		if ( ++i == argc ) error( "Provide a value for f." );
		arg->f = atoi( argv[i] ); 
		if ( arg->f < 0 ) error( "f must not be negative." );
		break;
	    case 'r': 
		if ( ++i == argc ) error( "Provide a value for R." );
		arg->R = atoi( argv[i] ); 
		if ( arg->R <= 0 ) error( "R must be positive." );
		break;
	    case 's': 
		if ( ++i == argc ) error( "Name a serial port." );
		arg->serial = argv[i]; 
		break;
	    case 'v':
		arg->version = 1;
		if ( ++i < argc )
		{
		    arg->version = 2;
		    arg->serial = argv[i]; 
		}
		break;
	    default: 
		error( "Unknown option." );
	}
    }
}

/**
 * Hauptprogramm
 */
int main( int argc, char ** argv ) 
{
    t_args arg;
    t_scope_data sd;
    t_signal signal;
    t_DUT dut;

    read_args( argc, argv, &arg );
    if ( arg.help ) help();
    if ( arg.version == 1 ) copyleft();
    if ( arg.version == 2 ) version( arg.serial );
    if ( arg.interactive ) 
    {
	interactive( &arg );
    }
    else
    {
	sd = measure( arg.serial, arg.autoscale );
	if ( sd.error[0] ) 
	    error( sd.error );
	signal = extract_signal( sd, arg.f );
	if ( signal.error[0] ) 
	    error( signal.error );
	dut = signal_to_dut( signal, arg.R, arg._r_p, arg._z_p );
	if ( arg.printing == 'H' )
	    print_impedance_human( dut );
	else
	    print_impedance_exp( dut );
    }
    return 0;
}


/* vim: set fdm=expr foldexpr=getline(v\:lnum)=~'^{'?'>1'\:getline(v\:lnum)=~'^}'?'<1'\:'-1' : */
