/**
 * osc_api -  library to communicate with HAMEG digital storage oscilloscopes
 *            via serial port.
 *         
 * Copyright (C) 2007, 2015 Lars Stollenwerk
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General
 * Public License along with this program; if not, write
 * to the Free Software Foundation, Inc., 51 Franklin St,
 * Fifth Floor, Boston, MA 02110, USA
 *
 * Contact the author via
 * formica@opppf.de
 * http://opppf.de/
 */
#ifndef OSC_API_H
#define OSC_API_H

#include <limits.h>

/**
 * write all errors to stdout
 */
#ifdef OSC_NOISY
#  define OSC_SHOUT fprintf( stderr, "%s\n", o->error )
#else
#  define OSC_SHOUT
#endif


/* development phase */

/* not (yet) implemented:
 * cursor usage 
 * display functions
 * switch functions */

/* ----------------------------------------------------------------------
				    types
 ---------------------------------------------------------------------- */


#if UCHAR_MAX != 255
#  error Expect char to be 8 bit!
#endif
#if USHRT_MAX != 65535
#  error Expect short int to be 16 bit!
#endif


/** field in ddf */
typedef unsigned char ddft;


/**
 * Device data field DDF
 */
typedef struct
{
  ddft ch1, ch2, vermode, tba, tbb, hormode, trig, strmode, 
       ch2var, ch1var, trsep, hold_off, intens_a, int_b;
} OSC_DDF;


/**
 * Device data field DDF1.
 * Values are 10 Bit unsigned: [0x000 .. 0x3ff]
 */
typedef struct
{
  short int tr_a_lev;
  short int tb_a_var;
  short int xpos;  /* FC >= 1.1: 16 bit signed, 1000 LSB/DIV */
  short int y2pos; /* FC >= 1.1: 16 bit signed, 1000 LSB/DIV */
  short int y1pos; /* FC >= 1.1: 16 bit signed, 1000 LSB/DIV */
  short int tr_b_lev;
  short int tb_b_var;
  short int del_pos; /* 12 bit?;  FC >= 2.0:  16 bit signed, 1000 LSB/DIV */
} OSC_DDF1;


/* bits in ch1, ch2 */
#define OSC_CHx_GND	0x80
#define OSC_CHx_AC 	0x40
#define OSC_CHx_INV	0x20
#define OSC_CHx_ON 	0x10
#define OSC_CHx_YS 	0x0f
/* bits in vermode */
#define OSC_VER_ALTTRG	0x80
#define OSC_VER_TK1	0x40 /* since version 2.00 */
#define OSC_VER_BWL	0x20 /* HM 200x only */
#define OSC_VER_CHOP	0x10
#define OSC_VER_ADD	0x08
#define OSC_VER_TK2	0x04 /* since version 2.00 */
#define OSC_VER_TRS	0x03
#define OSC_VER_TRS_CH1	0x00 /* value in OSC_VER_TRS */
#define OSC_VER_TRS_CH2	0x01 /* value in OSC_VER_TRS */
#define OSC_VER_TRS_EXT	0x02 /* value in OSC_VER_TRS */
/* bits in tba */
#define OSC_TBA_ZIN	0x80 /* HM40x since FC 2.09 and HM200x */
#define OSC_TBA_ANA	0x20 /* since version 2.02 */
#define OSC_TBx_TS	0x1f
/* bits in tbb */
#define OSC_TBB_NFL	0x80 /* triggr on negative flank */
#define OSC_TBB_BTR	0x40 /* wait for trigger after delay */
/* define OSC_TBx_TS	0x1f    same as tba */
/* bits in hormode */
#define OSC_HOR_CT	0x80
#define OSC_HOR_XY	0x40
#define OSC_HOR_x10	0x20
#define OSC_HOR_STORE	0x10
#define OSC_HOR_PP	0x08 /* only HM1207 since 3.24 */
#define OSC_HOR_TB	0x07
#define OSC_HOR_TB_a	0x00 /* value in OSC_HOR_TB */
#define OSC_HOR_TB_ab	0x01 /* value in OSC_HOR_TB */
#define OSC_HOR_TB_alt	0x02 /* value in OSC_HOR_TB */
#define OSC_HOR_TB_b	0x03 /* value in OSC_HOR_TB */
/* bits in trig */
#define OSC_TRI_NFL	0x80
#define OSC_TRI_PP	0x20
#define OSC_TRI_NORM	0x10
#define OSC_TRI_COUPL	0x07
#define OSC_TRI_COUPL_AC	0x00 /* value in OSC_TRI_COUPL */
#define OSC_TRI_COUPL_DC	0x01 /* value in OSC_TRI_COUPL */
#define OSC_TRI_COUPL_HF	0x02 /* value in OSC_TRI_COUPL */
#define OSC_TRI_COUPL_NR        0x03 /* -"-, not in 305, 404, 407 */
#define OSC_TRI_COUPL_LF	0x04 /* -"-,  -1 in 305, 404, 407 */
#define OSC_TRI_COUPL_		     /* other depend on hardware */
/* bits in strmode */
#define OSC_STR_REF1	0x80
#define OSC_STR_REF2	0x40
#define OSC_STR_PRET	0x38
#define OSC_STR_PRET_m75	0x00 /* value in OSC_STR_PRET */
#define OSC_STR_PRET_m50	0x08 /* value in OSC_STR_PRET */
#define OSC_STR_PRET_m25	0x10 /* value in OSC_STR_PRET */
#define OSC_STR_PRET_0		0x18 /* value in OSC_STR_PRET */
#define OSC_STR_PRET_25 	0x20 /* value in OSC_STR_PRET */
#define OSC_STR_PRET_50 	0x28 /* value in OSC_STR_PRET */
#define OSC_STR_PRET_75 	0x30 /* value in OSC_STR_PRET */
#define OSC_STR_PRET_100 	0x38 /* value in OSC_STR_PRET */
#define OSC_STR_MODE	0x07
#define OSC_STR_MODE_RFR	0x00 /* value in OSC_STR_MODE */
#define OSC_STR_MODE_SGL	0x01 /* value in OSC_STR_MODE, not vers 2.02 */
#define OSC_STR_MODE_ROL	0x02 /* value in OSC_STR_MODE */
#define OSC_STR_MODE_ENV	0x03 /* value in OSC_STR_MODE */
#define OSC_STR_MODE_AVM	0x04 /* value in OSC_STR_MODE */


/** maximum length of error string */
#define OSC_ERROR_STR_LEN 512

/**
 * scope handle.
 * Holds iscope state and all information for communication.
 */
typedef struct
{
  int fd;			/* file descriptor serial port */
  char * tty;			/* file name for serial port */
  char version[15];		/* version, e.g. 'FC1.01 DG1.03' */
  char id[27];			/* id,      e.g. 'HM1507    V01  MB01  DG01' */
  float fc;			/* fc number in version */
  int model;			/* model HMxxxx from id */
  char error[OSC_ERROR_STR_LEN];/* Error description as 0 terminated string,
                                   length 0 means no error */
  OSC_DDF ddf;
  OSC_DDF1 ddf1;
} OSC;



/* ----------------------------------------------------------------------
			     low level functions
 Basic communication. 
 ---------------------------------------------------------------------- */


/**
 * Open connection to scope.
 *
 * c	   filename of serial port
 * return: OSC descriptor
 *         possible error in field error.
 */
OSC * osc_open( const char * c );


/**
 * Close Connection to scope.
 *
 * o	OSC struct
 */
void osc_close( OSC * o );


/**
 * Write 0 terminated string to scope.
 *
 * o	OSC struct
 * c	string to write
 * return: 0 = success; -1 = error
 */
int osc_write_str( OSC * o, const char * c );


/**
 * Write string with n bytes to scope.
 *
 * o	OSC struct
 * c	string to write
 * n	length of string
 * return: 0 = success; -1 = error
 */
int osc_write_strn( OSC * o, const char * c, int n );


/**
 * Read answer from scope.
 * Answer comprises string c and m further bytes, written to b.
 *
 * o	OSC struct
 * c	string to receive
 * b	destination for answer
 * m	bytes of answer
 * return: 0 = success; -1 = error
 */
int osc_read_str( OSC * o, const char * c, void * b, int m );


/**
 * Read answer from scope.
 * Answer comprises string c, length n, and m further bytes, written to b.
 *
 * o	OSC struct
 * c	string to receive
 * b	destination for answer
 * m	bytes of answer
 * return: 0 = success; -1 = error
 */
int osc_read_strn( OSC * o, const char * c, int n, void * b, int m );


/**
 * Read returncode R CR NL
 *
 * o	OSC struct
 * return: The R, -1 on error
 */
int osc_get_returncode( OSC * o );



/**
 * Check whether returncode is '0' = ok.
 * A correct communication with returncode != '0' is considered as error.
 * 
 * s	additional text, e. g. function name.
 * o	OSC struct
 * return: 0  = success, -1 = error;
 */
int osc_check_returncode( OSC * o, const char * s );


/* ----------------------------------------------------------------------
			     high level functions
 Complete communication for a task.
 No further check of returncode needed. 
 ---------------------------------------------------------------------- */


/**
 * Read scope state. 
 * Is automaticcaly done on open.
 *
 * o	OSC struct
 * return: 0 = success; -1 = error
 */
int osc_read_state( OSC * o );


/**
 * Perform autoset.
 *
 * o	OSC struct
 * return: 0 = success; -1 = error
 */
int osc_set_auto( OSC * o );


/**
 * bell
 *
 * o	OSC struct
 * b	bell type to ring
 * return: 0 = success; -1 = error
 */
int osc_bell( OSC * o, int b );

#define OSC_BELL_KEYOK	'0' /* Key ok tone */
#define OSC_BELL_KEYERR	'1' /* Key error tone */
#define OSC_BELL_ERROR	'2' /* long tone */
#define OSC_BELL_2	'3' /* 2 short tones */
#define OSC_BELL_3	'4' /* 3 short tones */
#define OSC_BELL_6	'5' /* 6 short tones */


/**
 * Was scope used manually?
 * Since FC 1.1
 *
 * o    	OSC struct
 * return: 1 = yes, 0 = no, -1 = error
 */
int osc_changed_manually( OSC * o );


/* not implemented: LK
 * local lock
 */

/* not implemented: QUICKST
 * quicstart
 */

/* --- data acquisition ------------------------------------------------- */

/**
 * Set stormode.
 * mode is one of OSC_STR_MODE_..  RFR, SGL, ROL, ENV, AVM
 *
 * o            scope handle
 * mode		desired mode
 * return: 0 = success, -1 = error
 */
int osc_set_stormode( OSC * o, int mode );


/**
 * Get storemode.
 *
 * o            scope handle
 * return: one of OSC_STR_MODE_..  SGL, ROL, ENV, AVM
 */
#define osc_get_stormode(o) ( (o)->ddf.strmode & OSC_STR_MODE )


/**
 * Read data from cahannel n.
 * Values in Volt.
 * Read beginns at offset os and is len bytes long.
 * read whole channel: os = 0, len = 2048
 *
 * o		OSC struct
 * ch		channel number 1, 2
 * os		offset
 * len		length
 * b		buffer to write to
 * return: 0 = success; -1 = error; -2 = clip
 */
int osc_read_chan( OSC * o, int ch, int os, int len, double * b );


/**
 * set average number for acquisition
 *
 * o	OSC struct
 * e	exponent: number is 2^e; e in [1..9]
 * return: 0 = success; -1 = error
 */
int osc_set_avrnm( OSC * o, int e );


/**
 * get average number for acquisition.
 * number is 2^e; e in [0..9].
 *
 * o	OSC struct
 * return: e, -1 = error
 */
int osc_get_avrnm( OSC * o );


/* not implemnted: WFMPRE
 * additional information about waveform
 */

/* --- Y settings ------------------------------------------------------- */

/**
 * Set channel settings.
 * state is a sum of OSC_CHx_GND, OSC_CHx_AC, OSC_CHx_INV, OSC_CHx_ON
 *
 * o    	OSC struct
 * ch		channel, 1 or 2
 * state	desired state, -1 = don't change
 * ys		y-scale as index in OSC_V_DIV, -1 = don't change
 * var  	varscale, 0..0xff, 0xff = claibrated, -1 = don't change
 * return: 0 = success; -1 = error
 */
int osc_set_ch( OSC * o, int ch, int state, int ys, int var );

#define OSC_V_DIVN 14

/** Y amplification in DDF to voltage scale in V/DIV */
extern const double OSC_V_DIV[OSC_V_DIVN];


/**
 * voltage scale to index in OSC_V_DIV
 *
 * vs	voltage scale in V/DIV
 * return: corresponding index in OSC_V_DIV, -1 on error.
 */
int osc_vs_to_int( double vs );


/**
 * Get channel state
 * state is a sum of OSC_CHx_GND, OSC_CHx_AC, OSC_CHx_INV, OSC_CHx_ON
 *
 * o    	OSC struct
 * ch		channel, 1 or 2
 * return: state
 */
int osc_get_ch_state( OSC * o, int ch );

/** Test if channel is on.  */
#define osc_ch_on(o,x) ((o)->ddf.ch##x & OSC_CHx_ON)
/** Test if channel is inversed */
#define osc_ch_inv(o,x) ( (o)->ddf.ch##x & OSC_CHx_INV)
/** Test if channel is AC coupled */
#define osc_ch_ac(o,x) ((o)->ddf.ch##x & OSC_CHx_AC)
/** Test if channel is grounded */
#define osc_ch_gnd(o,x) ( (o)->ddf.ch##x & OSC_CHx_GND)


/**
 * Get voltage scale.
 *
 * o    	OSC struct
 * var		if not NULL, var setting is written here
 * ch		channel, 1 or 2
 * return: voltage scale as index in OSC_V_DIV
 */
int osc_get_ch_scale( OSC * o, int ch, int * var );


/**
 * Set y position on screen.
 * The value is an approximation. The middle may differ for both channels
 * and may be approx. 1 DIV off the screen center.
 *
 * o            OSC struct
 * ch 		channel, 1 or 2
 * pos		-1 = bottom, 0 = middle, 1 = top
 * return: 0 = success, -1 = error
 */
int osc_set_ypos( OSC * o, int ch, double pos );


/**
 * Get y position on screen
 *
 * o            OSC struct
 * ch 		channel, 1 or 2
 * return: -1 = bottom, 0 = middle, 1 = top
 */
double osc_get_ypos( OSC * o, int ch );


/* --- horizontal mode -------------------------------------------------- */

/**
 * Set horizontal mode.
 * Mode comprises OSC_HOR_... CT, XY, x10, STORE, PP, TB_a, TB_ab, TB_alt, TB_b
 * Mask comprises OSC_HOR_... CT, XY, x10, STORE, PP, TB
 * Only features named in mask are changed. They get the value given in mode.
 *
 * o    	OSC struct
 * mode		desired mode
 * mask		desired mask
 * return: 0 = success, -1 = error
 */
int osc_set_hormode( OSC * o, int mode, int mask );


/**
 * Get timebase.
 * Timebase settings as in osc_set_hormode.
 * return: one of TB_a, TB_ab, TB_alt, TB_b
 */
#define osc_get_timebase(o) ( (o)->ddf.hormode & OSC_HOR_TB )

/** Component test on */
#define osc_is_ct(o) ( (o)->ddf.hormode & OSC_HOR_CT )
/** XY mode on */
#define osc_is_xy(o) ( (o)->ddf.hormode & OSC_HOR_XY )
/** x10 stretch on */
#define osc_is_x10 ( (o)->ddf.hormode & OSC_HOR_x10 )
/** store mode on */
#define osc_is_store(o) ( (o)->ddf.hormode & OSC_HOR_STORE )
/** PP detect on */
#define osc_is_pp(o) ( (o)->ddf.hormode & OSC_HOR_PP )
 
/* --- timebase settings ------------------------------------------------ */

/**
 * Set timebase.
 * state for TB A is sum of: OSC_TBA_ZIN, OSC_TBA_ANA
 * state for TB B is sum of: OSC_TBB_NFL, OSC_TBB_BTR
 *
 * o		OSC struct
 * tb		timebas, 'A' or 'B'
 * state	desired state, -1 = don't schange
 * ts		timescale as index in OSC_s_DIV, -1 = don't change
 * var		varscale, 0..0x3ff, 0 = claibrated, -1 = dont't change
 * return: 0 = success; -1 = error
 */
int osc_set_tb( OSC * o, int tb, int state, int ts, int var );


#define OSC_s_DIVN 0x1D

/** Timebase in DDF to timescale in s/DIV  */
extern const double OSC_s_DIV[OSC_s_DIVN];


/**
 * timescale to index in OSC_s_DIV
 *
 * ts	timescale in s/DIV
 * return: corresponding index in OSC_s_DIV, -1 on error.
 */
int osc_ts_to_int( double ts );


/**
 * Get timebasestate
 * state is sum of:
 * TB A: OSC_TBA_ZIN, OSC_TBA_ANA
 * TB B: OSC_TBB_NFL, OSC_TBB_BTR
 *
 * o            OSC struct
 * tb		timebase 'A' or 'B'
 * return: state
 */
int osc_get_tb_state( OSC * o, int tb );


/**
 * Get time scale
 *
 * o            OSC struct
 * var          if not NULL, var setting is written here
 * b            timebase 'A' or 'B'
 * return: time scale as index in OSC_s_DIV
 */
int osc_get_tb_scale( OSC * o, int tb, int * var );


/**
 * Set x position on screen
 *
 * pos		-1 = very left, 0 = middle, 1 = very right
 * o            OSC struct
 * return: 0 = success, -1 = error
 */
int osc_set_xpos( OSC * o, double pos );


/**
 * Get x position on screen
 *
 * o            OSC struct
 * return: -1 = very left, 0 = middle, 1 = very right
 */
double osc_get_xpos( OSC * o );


/* --- vertical mode ---------------------------------------------------- */

/**
 * Set vertical mode.
 * Mode comprises OSC_VER_..  ALTTRG, TK1, BWL, CHOP, ADD, TK2, TRS_CH1,
 *                                                              TRS_CH2, TRS_EXT
 * Mask comprises OSC_VER_..  ALTTRG, TK1, BWL, CHOP, ADD, TK2, TRS
 *
 * o            OSC struct
 * mode         desired mode
 * mask         desired mask
 * return: 0 = success, -1 = error
 */
int osc_set_vermode( OSC * o, int mode, int mask );


/**
 * Get trigger source.
 * Trigger source as in osc_set_vermode.
 * return: one of OSC_VER_.. TRS_CH1, TRS_CH2, TRS_EXT
 */
#define osc_get_trigger_source(o) ( (o)->ddf.vermode & OSC_VER_TRS )


/** Alternating trigger on */
#define osc_is_alt_trg(o) ( (o)->ddf.vermode & OSC_VER_ALTTRG )
/** Is chopping on */
#define osc_is_chop(o) ( (o)->ddf.vermode & OSC_VER_CHOP )
/** Is channel add on */
#define osc_is_ch_add(o) ( (o)->ddf.vermode & OSC_VER_ADD )
/** Is BWL on */
#define osc_is_bwl(o) ( (o)->ddf.vermode & OSC_VER_BWL )
/** CH1 probe 1:10 */
#define osc_is_ch1_x10(o) ( (o)->ddf.vermode & OSC_VER_TK1 )
/** CH2 probe 1:10 */
#define osc_is_ch2_x10(o) ( (o)->ddf.vermode & OSC_VER_TK2 )

/* --- trigger settings ------------------------------------------------- */

/**
 * Set Holdoff time
 *
 * o    	OSC struct
 * t		holdoff time, 0x00 is shortest, 0xff is longest
 * return: 0 = success; -1 = error
 */
int osc_set_holdoff( OSC * o, int t );


/**
 * Get Holdoff time
 * 0x00 is shortest, 0xff is longest
 *
 * o    	OSC struct
 * return: holdoff time
 */
#define osc_get_holdoff(o) ( (o)->ddf.hold_off )


/**
 * Reset trigger.
 * Single mode storage: reset and wait for trigger.
 * Storage: Restart acquisition now.
 * Analog: start trigger now.
 *
 * o		OSC struct
 * return: 0 = success; -1 = error
 */
int osc_reset( OSC * o );


/**
 * set trigger mode.
 * mode comprises: OSC_TRI_...        NFL, PP, NORM
 * coupling:       OSC_TRI_COUPL_...  AC, DC, HF
 *
 * o		OSC struct
 * mode		desired mode, -1 = don't change
 * coupl	coupling, -1 = don't change
 * return: 0 = success; -1 = error
 */
int osc_set_trigger( OSC * o, int mode, int coupl );


/**
 * Get trgger mode.
 *
 * o		OSC struct
 * return: one of OSC_TRI_...        NFL, PP, NORM
 */
#define osc_get_trig_mode ( (o)->ddf.trig & ~OSC_TRI_COUPL )


/**
 * Get trgger coupling.
 *
 * o		OSC struct
 * return: one of OSC_TRI_COUPL_...  AC, DC, HF
 */
#define osc_get_trig_src ( (o)->ddf.trig & OSC_TRI_COUPL )


/**
 * Find out whether scope triggers.
 * States are:
 *   TR_YES  scope triggers
 *   TR_NO   socee triggers not
 *   TR_WAIT scope in ingle modus and waits or acquisition not complete
 *   -1	     error
 *
 * o		scope handle
 * fast		1 = without use of 200 ms monoflop
 * return: trigger status 
 */
int osc_get_trgsta( OSC * o, int fast );

#define TR_NO 0
#define TR_YES 1
#define TR_WAIT 2


/**
 * Set pretrigger
 *
 * pt	one of OSC_STR_PRET_..  m75, m50, m25, 0, 25, 50, 75, 100
 * o		scope handle
 * return: 0 = success, -1 = error
 */
int osc_set_pretrig( OSC * o, int pt );


/**
 * Get pretrigger.
 *
 * o		scope handle
 * return: one of OSC_STR_PRET_..  m75, m50, m25, 0, 25, 50, 75, 100
 */
#define osc_get_pretrig(o) ( (o)->ddf.strmode & OSC_STR_PRET )


/* not implemented: TRGPOS
 * trigger position, since FC 3.24
 */

/* not implemented: TRGVAL
 * values describing trigger level
 *
 * not implemented: TRGLEV
 * trigger level
 */

/* not implemented: TRSEP
 * trace separation for second time base
 */

/* Not implemented: DELPOS
 * Set / Get delay position.
 * 
 * What are the allowed values?
 * Which scopes use it?
 * Is it delay for timebase B?
 */

/* --- reference wave --------------------------------------------------- */

/**
 * Not implemented.
 * Read reference waveform n.
 * Read beginns at offset os and is len bytes long.
 * read whole channel: os = 0, len = 2048
 *
 * o		OSC struct
 * n		reference number 1, 2
 * os		offset
 * len		length
 * b		buffer to write to
 * return: 0 = success; -1 = error
 */


/* not implemented:
 * switchin on and off reference waves
 * to be done in DDF STRMODE.
 */

/* not implemented: 
 * REFxPOS	position of reference wave
 * REFxPRE	additional info fro reference wave
 * RREFPRE	set scope to state that was used foe reference wave
 * SAVEREF	save wave as reference wave
 */

/* not implemented: 
 * RECDF	recall ddf
 * SAVEDF	save ddf
 */

/* not implemented: WRREF
 * write wave in reference memory
 */

/* --- Hold waveform ---------------------------------------------------- */

/**
 * Set hold waveform.
 *
 * o    	OSC struct
 * x		1 = hold, 0 = no hold
 * return: 0 = success, -1 = error
 */
int osc_set_hold( OSC * o, int x );


/**
 * Get hold waveform.
 *
 * o    	OSC struct
 * return: 1 = hold, 0 = no hold, -1 = error
 */
int osc_get_hold( OSC * o );

/* Not implemented: HLDxPOS
 * Shift of holded waveforms
 */


#endif
